/*==========================================================================

----------------------
--- Licensing Info ---
----------------------

Copyright (C) 2023 by 1xAero

This program is free software: you can redistribute it and/or modify
it, share, upload wherever you want without author's permission.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

-------------------
--- Description ---
-------------------

This script automatically spawns and respawns monster_apache and some small monsters. 
Also you can spawn any monster manually. Don't forget to precache their resources.
Currently supported mods:

   - Half-Life
		
--------------------
--- Requirements ---
--------------------

* Mods: Half-Life 25th Anniversary Update
* Metamod 1.21p38: 
  - https://metamod-p.sourceforge.net/
* AMXX Version 1.10: 
  - https://www.amxmodx.org/downloads-new.php?branch=master

-----------------
--- Changelog ---
-----------------
	
* release: 1.0 (Dec 21 2023)
  - First public release.

* release: 1.1 (Dec 24 2023)
  - Fixed crash on monsters spawn and death. Added hassassin, zombie, agrunt, alien slave,
  alien controller, removed hgrunt for now. 

* release: 1.2 (Dec 27 2023)
  - Bugfixes and Quality of Life: Fixed game crashing when the monsters limit exceeded 
  the specified amount and the monsters deletion script was launched.
  Monsters now will not disappear instantly after reaching limit. 
  Added: bullsquid, hgrunt, headcrab, boss gargantua.

* release: 1.3 (Dec 27 2023)
  - Fixed pev_flags -> pev_spawnflags for SF_MONSTER_FALL_TO_GROUND.
  Removed additional gargantua checks.

* release: 1.4 (Dec 27 2023)
  - Fixed crashes on some maps like crossfire, where weapon_9mmAR is spawning by default.
  Added checks for hgrunt weapon_9mmAR.  

===========================================================================*/

#include <amxmodx>
#include <fakemeta>
#include <hamsandwich>

#include "monsters/0x0001_setup.sma"
#include "monsters/0x0002_basefunctions.sma"

#define PLUGIN "Half-Life: Monsters"
#define VERSION "1.4"
#define AUTHOR "1xAero"	

public plugin_init()
{
	register_plugin(PLUGIN, VERSION, AUTHOR);

	//Force allowmonster in multiplayer.
	set_cvar_num(ALLOWMONSTERS, TRUE);

	//Hook hgrunt weapon.	
	register_forward(FM_Touch, "Forward_CBaseGameTouch", FALSE);	
	
	//Hook entities from Half-Life	
	register_forward(FM_Think, "Forward_CBaseGameThink", FALSE);

	//Spawn small monsters on victims origin.
	RegisterHam(Ham_Killed, "player", "HamF_Killed_Post", TRUE);

	//Manually spawn monsters.
	register_clcmd("_monster", "Function_CBaseMonsterSpawn", FCVAR_PROTECTED);

	//Monster apache not available to spawn on some maps due to map size.
	if(!Function_RestrictedMaps())
	{
		//Start apache task.
		remove_task(TASK_MONSTERSETUP);
		set_task(GLB_TIMER, "Function_SetupMonster", TASK_MONSTERSETUP);
	}
}

//Remove his weapon on drop, this is not Player weapon or worldspawned.
public Forward_CBaseGameTouch(iEntity, iTouched)
{
	if(!pev_valid(iEntity))
		return FMRES_IGNORED;

	//Pass only for weapon_9mmAR.
	if(!equal(Function_GetEntityClassname(iEntity), weapon_hgrunt_classname))
		return FMRES_IGNORED;			

	/*How to detect his weapon and not to remove default 9mmAR from the map? 
	The owner not exists because the weapon are just spawning after hgrunt death, but in hgrunt.cpp line 297/298 we have the code:
	pGun->pev->velocity = Vector(RANDOM_FLOAT(-100,100), RANDOM_FLOAT(-100,100), RANDOM_FLOAT(200,300));
	pGun->pev->avelocity = Vector(0, RANDOM_FLOAT(200, 400), 0);*/

	//Do this checks.
	new Float: iVelocity[3], Float: iAvelocity[3];

	pev(iEntity, pev_velocity, iVelocity);
	pev(iEntity, pev_avelocity, iAvelocity);

	//Got it.
	new Float: iDropSpeed = floatsqroot(iVelocity[0] * iVelocity[0] + iVelocity[1] * iVelocity[1] + iVelocity[2] * iVelocity[2]);
	new Float: iRotateSpeed = floatsqroot(iAvelocity[0] * iAvelocity[0] + iAvelocity[1] * iAvelocity[1] + iAvelocity[2] * iAvelocity[2]);

	//Our weapon.
	if(iDropSpeed > 0.0 && iRotateSpeed > 0.0)
	{
		set_pev(iEntity, pev_flags, FL_KILLME);
		dllfunc(DLLFunc_Think, iEntity);
	}

	return FMRES_IGNORED;					
}

//Entities think
public Forward_CBaseGameThink(iEntity)
{
	if(!pev_valid(iEntity))
		return FMRES_IGNORED;

	//Do some stuff with ground monsters and pass only for apache.
	if(!equal(Function_GetEntityClassname(iEntity), monster_apache_classname))
		return FMRES_IGNORED;	
		
	//Not killed yet.	
	if(pev(iEntity, pev_takedamage) != DAMAGE_NO)
	{

		//Only rocket attack
		#if defined MONSTER_APACHE_ROCKET_ONLY

			set_ent_data(iEntity, "CApache", "m_iRockets", APACHE_HEAVY);

		//Both
		#else

			//Let it attack with rockets and firegun
			if(c_monster_apache_think < get_gametime())
				c_monster_apache_think = get_gametime() + APACHE_ROCKET;					

			//Can use rocket attack.		
			if(c_monster_apache_think > get_gametime() + APACHE_FIREGUN)
				set_ent_data(iEntity, "CApache", "m_iRockets", APACHE_HEAVY);

			//Firegun attack.
			else
				set_ent_data(iEntity, "CApache", "m_iRockets", APACHE_LIGHT);

		#endif	

		//Dunno't allow apache move upwards.
		set_ent_data_float(iEntity, "CApache", "m_flForce", APACHE_POWER);
	
		//TODO hook this.
		/*set_ent_data_float(iEntity, "CApache", "m_flNextRocket", get_gametime());
		set_ent_data_entity(iEntity, "CBaseEntity", "m_pGoalEnt", NULL);		
		set_ent_data_entity(iEntity, "CBaseMonster", "m_hEnemy", NULL);*/	

		//Could make here lots of hardcoding stuff...
	}	
		
	return FMRES_IGNORED;	
}

//Spawn at death.
public HamF_Killed_Post(iVictim, iKiller, Gib)
{
	if(!is_user_connected(iVictim)) 
		return;

	//Explode all monsters if limit is reached and continue spawn.
	if(Function_GetAllMonsters() > MONSTER_AMOUNT)	
		Function_KillAllMonsters();		

	new Float: iOrigin[3];
	Function_SetMonsterSpawn(iVictim, iOrigin);

	//Spawn at victim's origin + z
	if(pev(iVictim, pev_waterlevel) == NULL && Function_GetAllMonsters() < MONSTER_AMOUNT + 1)
		Function_SpawnRandomMonster(iOrigin);	
}

//Spawn manually.
public Function_CBaseMonsterSpawn(iPlayer)
{
	//Kill monsters.
	if(Function_GetAllMonsters() > MONSTER_AMOUNT - 1)
	{
		Function_KillAllMonsters();

		return PLUGIN_HANDLED;
	}

	new Float: iOrigin[3];

	Function_SetMonsterSpawn(iPlayer, iOrigin);
	Function_SpawnRandomMonster(iOrigin);

	return PLUGIN_HANDLED;
}

//Global task for monster_apache.
public Function_SetupMonster(TASK_MONSTERSETUP)
{
	//Call this monster.	
	if(!Function_GetMonsterEntity(monster_apache_classname) && !Function_GetMonsterEntity(monster_gargantua_classname))
	{
		new iRandom;
		iRandom = random_num(1, 2);

		switch(iRandom)
		{
			case 1: Function_CBaseMonster(monster_apache_classname, MONSTER_APACHE_HEALTH);
			case 2: Function_CBaseMonster(monster_gargantua_classname, MONSTER_GARGANTUA_HEALTH);
		}
	}	
	
	remove_task(TASK_MONSTERSETUP);
	set_task(GLB_TIMER, "Function_SetupMonster", TASK_MONSTERSETUP);
}

//Remove dead bodies with a bit delay.
public Function_RemoveBodies(iCBaseMonster)
{
	//Only valid or just throws instant crash.
	if(Function_MonsterSafe(iCBaseMonster) && pev(iCBaseMonster, pev_deadflag) < DEAD_RESPAWNABLE)
	{	
		set_pev(iCBaseMonster, pev_flags, FL_KILLME);
		dllfunc(DLLFunc_Think, iCBaseMonster);
	}			
}

//Set normal monster condition
public Function_SetMonsterMovetype(CBaseMonster)
{
	if(Function_MonsterSafe(CBaseMonster) && pev(CBaseMonster, pev_deadflag) < DEAD_DYING)
	{
		//Set states.
		set_pev(CBaseMonster, pev_solid, SOLID_SLIDEBOX);
		set_pev(CBaseMonster, pev_takedamage, DAMAGE_YES);

		//Render default.
		SetRendering(CBaseMonster);

		//TODO: Recheck airstuck!
		UnstuckMonster(CBaseMonster, 128, 128);

		//Dunno't force for everyone, maybe not clipped?
		if(!engfunc(EngFunc_WalkMove, CBaseMonster,  0.0, 0, WALKMOVE_NORMAL))	
			engfunc(EngFunc_DropToFloor, CBaseMonster);		
	}	
}