//Call HL base monster.
stock Function_CBaseMonster(const Monster[], Float: iHealth, Float: iOrigin[3] = { 0.0, 0.0, 0.0 })
{
	//Create monster.
	new CBaseMonster = engfunc(EngFunc_CreateNamedEntity, engfunc(EngFunc_AllocString, Monster));

	if(!pev_valid(CBaseMonster))	
		return;

	//Origin.
	new Float: fOrigin[3];
	
	//Set origin of big manually.
	if(equal(Function_GetEntityClassname(CBaseMonster), monster_apache_classname) || equal(Function_GetEntityClassname(CBaseMonster), monster_gargantua_classname))
		engfunc(EngFunc_SetOrigin, CBaseMonster, Function_GetMonsterOrigin(fOrigin));

	//Let other allow to spawn via player cmd.
	else
		engfunc(EngFunc_SetOrigin, CBaseMonster, iOrigin);	

	//Spawn
	dllfunc(DLLFunc_Spawn, CBaseMonster);
	
	//Set health.
	set_pev(CBaseMonster, pev_health, iHealth);

	//Set key
	set_pev(CBaseMonster, pev_iuser1, MONSTER_BIND);
	
	/*Bypass weird ground checks by HL for small monsters and gargantua. We don't need that for multiplayer.
	monsters.cpp line 2086 - Raise monster off the floor one unit, then drop to floor*/
	if(!equal(Function_GetEntityClassname(CBaseMonster), monster_apache_classname))
	{
		//Give grenadelauncher and 9mmAR
		/*if(equal(Function_GetEntityClassname(CBaseMonster), monster_hgrunt_classname))	
			set_pev(CBaseMonster, pev_weapons, HGRUNT_9MMAR | HGRUNT_GRENADELAUNCHER);*/	

		//Dunno't block at spawn.
		set_pev(CBaseMonster, pev_solid, SOLID_NOT);
		set_pev(CBaseMonster, pev_takedamage, DAMAGE_NO);

		//Render
		SetRendering(CBaseMonster, kRenderFxDistort, NULL, NULL, NULL, kRenderTransAdd, 125);

		//Set to get rid of game basic setup.
		set_pev(CBaseMonster, pev_spawnflags, SF_MONSTER_FALL_TO_GROUND);

		//Hack movetype for all and drop them down.
		set_task(DOUBLE_TIMER, "Function_SetMonsterMovetype", CBaseMonster);		

		//Set next action delay.
		set_pev(CBaseMonster, pev_nextthink, get_gametime() + DOUBLE_TIMER);		
	}
}

//Force unstuck at spawn.
stock UnstuckMonster(iEntity, i_StartDistance, i_MaxAttempts)
{
	#define SKIP_IGNORE -1
	#define MIN_HEALTH 0.1

	enum Coord_e 
	{ 
		Float: x, 
		Float: y, 
		Float: z 
	}

	#define GetHullSize ((pev(iEntity, pev_flags) & FL_DUCKING) ? HULL_HEAD : HULL_HUMAN)	
	#define MAX_DISTANCE 1000

	//Not alive, ignore.
	if(pev(iEntity, pev_deadflag) == DEAD_DEAD)
		return SKIP_IGNORE;	
        
	static Float: vf_OriginalOrigin[Coord_e], Float: vf_NewOrigin[Coord_e];
	static i_Attempts, i_Distance;
        
	//Get the current player's origin.
	pev(iEntity, pev_origin, vf_OriginalOrigin);
        
	i_Distance = i_StartDistance;
        
	while(i_Distance < MAX_DISTANCE)
	{
		i_Attempts = i_MaxAttempts;
            
		while(i_Attempts --)
		{
			vf_NewOrigin[x] = random_float(vf_OriginalOrigin[x] - i_Distance, vf_OriginalOrigin[x] + i_Distance);
			vf_NewOrigin[y] = random_float(vf_OriginalOrigin[y] - i_Distance, vf_OriginalOrigin[y] + i_Distance);
			vf_NewOrigin[z] = random_float(vf_OriginalOrigin[z] - i_Distance, vf_OriginalOrigin[z] + i_Distance);
            
			engfunc(EngFunc_TraceHull, vf_NewOrigin, vf_NewOrigin, DONT_IGNORE_MONSTERS, GetHullSize, iEntity, NULL);
            
			//Free space found.
			if(get_tr2(NULL, TR_InOpen) && !get_tr2(NULL, TR_AllSolid) && !get_tr2(NULL, TR_StartSolid))
			{
				//Set the new origin.
				engfunc(EngFunc_SetOrigin, iEntity, vf_NewOrigin);
				
				return TRUE;
			}
		}
            
		i_Distance += i_StartDistance;
	}
        
	//Could not be found.
	return FALSE;
}

//Get amount of spawned things already.
stock Function_CBaseMonsterCount(const iClassname[])
{
	new iMonsters = NULL, iCBaseMonster = NULLENT;

	for(new CBaseMonsters = 0; ((iCBaseMonster = engfunc(EngFunc_FindEntityByString, iCBaseMonster, "classname", iClassname)) > NULL); CBaseMonsters ++)
		iMonsters ++;	
    
	//client_print(NULL, print_chat, "iCBaseMonster amount %d", iMonsters);

	return iMonsters;
}

//Get classname
stock Function_GetEntityClassname(iEntity)
{
	static iClassname[32];
	pev(iEntity, pev_classname, iClassname, charsmax(iClassname));

	return iClassname;
}

//Exists or not?
stock bool: Function_GetMonsterEntity(const iClassname[])
{
	new iEntity = NULLENT;

	if((iEntity = engfunc(EngFunc_FindEntityByString, iEntity, "classname", iClassname)) > NULL)
		return true;
		
	return false;	
}

//Return map spawn available.
stock bool: Function_RestrictedMaps()
{
	new Float: iOrigin[3];
	Function_GetMonsterOrigin(iOrigin);
	
	if(iOrigin[0] == 0.0 && iOrigin[1] == 0.0 && iOrigin[2] == 0.0)
		return true;
		
	return false;	
}

//Set origin.
stock Float: Function_SetMonsterSpawn(iPlayer, Float: iOrigin[3])
{
	pev(iPlayer, pev_origin, iOrigin);
	iOrigin[2] += NAN_ORIGIN;
}

//Kill all monsters and remove them.
stock Function_KillMonsters(const iClassname[])
{
	new iCBaseMonster = NULLENT;
    
	for(new CBaseMonsters = 0; ((iCBaseMonster = engfunc(EngFunc_FindEntityByString, iCBaseMonster, "classname", iClassname)) > NULL); CBaseMonsters ++)
	{
		//Gibs are also affecting on entity limit. Prevent MSG Overflow.
		if(pev_valid(iCBaseMonster) && pev(iCBaseMonster, pev_deadflag) < DEAD_RESPAWNABLE && pev(iCBaseMonster, pev_iuser1) == MONSTER_BIND)
		{	
			ExecuteHam(Ham_Killed, iCBaseMonster, iCBaseMonster, GIB_NEVER);
			set_task(GLB_TIMER * 2, "Function_RemoveBodies", iCBaseMonster);			
		}			
	}			
}

//Get mapname.
stock Function_GetMapName()
{
	new iMapName[32];
	get_mapname(iMapName, charsmax(iMapName));
	
	return iMapName;
}

//Spawn houndeye or hgrunt
stock Function_SpawnRandomMonster(Float: iOrigin[3])
{
	new iRandom;
	iRandom = random_num(1, 11);

	switch(iRandom)
	{
		case 1: Function_CBaseMonster(monster_houndeye_classname, MONSTER_HOUNDEYE_HEALTH, iOrigin);
		case 2: Function_CBaseMonster(monster_hassassin_classname, MONSTER_HASSASSIN_HEALTH, iOrigin);
		case 3: Function_CBaseMonster(monster_zombie_classname, MONSTER_ZOMBIE_HEALTH, iOrigin);
		case 4: Function_CBaseMonster(monster_agrunt_classname, MONSTER_AGRUNT_HEALTH, iOrigin);
		case 5: Function_CBaseMonster(monster_islave_classname, MONSTER_ISLAVE_HEALTH, iOrigin);
		case 6..8: Function_CBaseMonster(monster_hgrunt_classname, MONSTER_HGRUNT_HEALTH, iOrigin);
		case 9, 10: Function_CBaseMonster(monster_bullsquid_classname, MONSTER_BULLSQUID_HEALTH, iOrigin);
		case 11: Function_CBaseMonster(monster_headcrab_classname, MONSTER_HEADCRAB_HEALTH, iOrigin);
	}
}

//Kill all monsters
stock Function_KillAllMonsters()
{
	Function_KillMonsters(monster_houndeye_classname);
	Function_KillMonsters(monster_hassassin_classname);
	Function_KillMonsters(monster_zombie_classname);
	Function_KillMonsters(monster_agrunt_classname);
	Function_KillMonsters(monster_islave_classname);
	Function_KillMonsters(monster_hgrunt_classname);
	Function_KillMonsters(monster_bullsquid_classname);
	Function_KillMonsters(monster_headcrab_classname);
}

//Both monster type
stock Function_GetAllMonsters()
{
	new iTotal = NULL;
	iTotal = (Function_CBaseMonsterCount(monster_houndeye_classname) 
		PLUS Function_CBaseMonsterCount(monster_hassassin_classname) 
		PLUS Function_CBaseMonsterCount(monster_zombie_classname) 
		PLUS Function_CBaseMonsterCount(monster_agrunt_classname) 
		PLUS Function_CBaseMonsterCount(monster_islave_classname) 
		PLUS Function_CBaseMonsterCount(monster_hgrunt_classname) 
		PLUS Function_CBaseMonsterCount(monster_bullsquid_classname) 
		PLUS Function_CBaseMonsterCount(monster_headcrab_classname));
	
	//client_print(NULL, print_chat, "iCBaseMonster total %d", iTotal);

	return iTotal;
}

//Check if it's this monster.
stock bool: Function_MonsterSafe(iCBaseMonster)
{
	if(pev_valid(iCBaseMonster) && pev(iCBaseMonster, pev_iuser1) == MONSTER_BIND)
		return true;

	return false;
}

//RenderFx
stock SetRendering(iEntity, iFx = kRenderFxNone, r = 255, g = 255, b = 255, iRender = kRenderNormal, iAmount = 16)
{
	static Float: iColor[3];
	iColor[0] = float(r);
	iColor[1] = float(g);
	iColor[2] = float(b);

	set_pev(iEntity, pev_renderfx, iFx);
	set_pev(iEntity, pev_rendercolor, iColor);
	set_pev(iEntity, pev_rendermode, iRender);
	set_pev(iEntity, pev_renderamt, float(iAmount));
}

//Map spawns.
stock Float: Function_GetMonsterOrigin(Float: iOrigin[3])
{
	new iRandom;
	iRandom = random_num(1, 2);
	
	Function_GetMapName();

	if(equal(Function_GetMapName(), "boot_camp"))
		iOrigin = Float: { -817.0, 445.0, 463.0 };
		
	else if(equal(Function_GetMapName(), "bounce"))
	{
		switch(iRandom)
		{
			case 1: iOrigin = Float: { -143.0, 1027.0, -1800.0 };
			case 2: iOrigin = Float: { -226.0, -608.0, -1800.0 };
		}	
	}	
		
	else if(equal(Function_GetMapName(), "contamination"))
		iOrigin = Float: { 837.0, -116.0, 369.0 };

	else if(equal(Function_GetMapName(), "crossfire"))
		iOrigin = Float: { -156.0, 669.0, -1556.0 };

	else if(equal(Function_GetMapName(), "disposal"))
		iOrigin = Float: { 586.0, 735.0, 121.0 };

	else if(equal(Function_GetMapName(), "doublecross"))
		iOrigin = Float: { 324.0, -976.0, -1433.0 };

	else if(equal(Function_GetMapName(), "gasworks"))
		iOrigin = Float: { -826.0, -352.0, 421.0 };

	else if(equal(Function_GetMapName(), "pool_party"))
		iOrigin = Float: { -195.0, 48.0, -222.0 };

	else if(equal(Function_GetMapName(), "xen_dm"))
		iOrigin = Float: { 486.0, -1920.0, -1333.0 };
		
	else
		iOrigin = Float: { 0.0, 0.0, 0.0 };

	return iOrigin;		
}